﻿namespace App336
{
  using System.Collections.Generic;
  using System.Threading.Tasks;
  using Windows.UI.Xaml.Controls;
  using Windows.UI.Xaml.Navigation;
  using System;
  using System.Linq;
  using Windows.Storage;
  using Windows.UI.Xaml;
  using Windows.UI.Popups;
  public sealed partial class MainPage : Page
  {
    enum VisualStates
    {
      Default,
      Submitting,
      Recording
    }
    // Clearly, with a bit of work you could drive all this speech interaction
    // with a JSON/XML file and some kind of interpreter but, for now, I'm
    // just hand cranking the code.
    public MainPage()
    {
      this.InitializeComponent();
      this.data = new DisplayTextViewModel();
      this.DataContext = this.data;
      this.conversation = new Conversation(this.mediaElement);
      this.restClient = new OxfordSpeakerIdRestClient();
      this.audioRecorder = new AudioRecorder();         
    }
    void SwitchState(VisualStates state)
    {
      VisualStateManager.GoToState(this, state.ToString(), true);
    }
    protected async override void OnNavigatedTo(NavigationEventArgs e)
    {
      base.OnNavigatedTo(e);

      await this.StartConversationAsync();
    }
    async Task StartConversationAsync()
    {
      var command = await this
        .DisplaySayAsync("login or register", "would you like to login or register?")
        .WaitForCommandsAsync("login", "register");

      switch (command.Text)
      {
        case "login":
          await this.LoginAsync();
          break;
        case "register":
          await this.RegisterAsync();
          break;
        default:
          break;
      }
    }
    async Task LoginAsync()
    {
      await this.GetAccountNumberAsync();

      var guid = await AccountMap.GetGuidForAccountNumberAsync(accountNumber);

      if (!guid.HasValue)
      {
        await this.DisplaySayAsync("not found",
          "sorry, I don't know that account number");
      }
      else
      {
        await this.RecordSpeechToFileAsync(DEFAULT_RECORD_TIME);

        VerificationResult result = null;

        using (var stream = await this.recordingFile.OpenReadAsync())
        {
          await this.MakeOxfordCallAsync(
            async () =>
            {
              result = await this.restClient.VerifyAsync(guid.Value, stream);
            }
          );
        }
        if (result.Result == VerificationStatus.Accept)
        {
          await
            this.DisplaySayAsync("logged in", "you are now logged in to your account");
        }
        else
        {
          await this.DisplaySayAsync("denied", "sorry, that voice pattern didn't verify");
        }
      }
      await this.StartConversationAsync();
    }
    async Task RegisterAsync()
    {
      await this.GetAccountNumberAsync();

      await this.MakeProfileAsync();

      await this.EnrollAsync();
    }
    async Task RecordSpeechToFileAsync(float recordTime)
    {
      await this.PopulateVerificationPhrasesAsync();
      await this.CreateStorageFileAsync();

      await this.DisplaySayAsync(
        "repeat the phrase", "when I say 'go'")
        .PauseAsync(TimeSpan.FromMilliseconds(500))
        .SayAsync("please repeat the phrase displayed");

      this.data.DisplayText = this.verificationPhrases[0].Phrase;

      await this.conversation.SayAsync("go");

      this.SwitchState(VisualStates.Recording);

      await this.audioRecorder.StartRecordToFileAsync(this.recordingFile);

      await this.RunProgressLoopAsync(0.0f, recordTime,
        TimeSpan.FromSeconds(recordTime));

      await this.audioRecorder.StopRecordAsync();

      this.SwitchState(VisualStates.Default);
    }
    async Task EnrollAsync()
    {
      var enrolled = false;

      while (!enrolled)
      {
        await this.RecordSpeechToFileAsync(DEFAULT_RECORD_TIME);

        this.data.DisplayText = "submitting";

        EnrollmentResult result = null;

        using (var stream = await this.recordingFile.OpenReadAsync())
        {
          await this.MakeOxfordCallAsync(
            async () =>
            {
              result = await this.restClient.EnrollAsync(
                this.verificationProfile, stream);
            }
          );
        }
        enrolled = result.EnrollmentStatus == EnrollmentStatus.Enrolled;

        if (!enrolled)
        {
          await this.conversation.SayAsync(
            $"I'll need you repeat that just {result.RemainingEnrollments} more time");
        }
        else
        {
          await AccountMap.SetGuidForAccountNumberAsync(this.accountNumber,
            this.verificationProfile.VerificationProfileId);
        }
      }
      await this.DisplaySayAsync("done", "ok, you're enrolled");
      await this.StartConversationAsync();
    }
    async Task RunProgressLoopAsync(float minimum, float maximum, TimeSpan time)
    {
      this.data.ProgressMinimum = minimum;
      this.data.ProgressMaximum = maximum;
      this.data.ProgressValue = 0;
      var totalSeconds = time.TotalSeconds;
      var increment = (maximum - minimum) / totalSeconds;

      for (int i = 0; i < time.TotalSeconds; i++)
      {
        this.data.ProgressValue += (float)increment;
        await Task.Delay(TimeSpan.FromSeconds(1));
      }
      this.data.ProgressValue = 0;
    }
    async Task CreateStorageFileAsync()
    {
      this.recordingFile = await ApplicationData.Current.TemporaryFolder.CreateFileAsync(
        RECORDING_FILE, CreationCollisionOption.ReplaceExisting);
    }
    async Task MakeOxfordCallAsync(Func<Task> call)
    {
      this.SwitchState(VisualStates.Submitting);

      try
      {
        await call();
      }
      catch (Exception ex)
      {
        var dialog = new MessageDialog(ex.Message, "failed, not likely to recover");
        await dialog.ShowAsync();
      }
      this.SwitchState(VisualStates.Default);
    }
    async Task PopulateVerificationPhrasesAsync()
    {
      if (this.verificationPhrases == null)
      {
        await this.MakeOxfordCallAsync(
          async () =>
          {
            var enumerable = await this.restClient.GetVerificationPhrasesAsync();
            this.verificationPhrases = enumerable.ToList();
          }
        );
      }
    }
    async Task MakeProfileAsync()
    {
      this.restClient = new OxfordSpeakerIdRestClient();

      await this.DisplaySayAsync(
        "making a new online profile for this account");

      await this.MakeOxfordCallAsync(
        async () =>
        {
          this.verificationProfile = await this.restClient.AddVerificationProfileAsync();
        }
      );
      this.data.DisplayText = "done";
    }
    void HandleError(Exception ex)
    {
      throw new NotImplementedException();
    }
    async Task GetAccountNumberAsync()
    {
      // We don't validate.
      bool done = false;
      string account = string.Empty;
      this.accountNumber = string.Empty;

      while (!done)
      {
        var digits = await this.DisplaySayAsync(
          "please say your account number")
          .WaitForDigitsAsync();

        account = digits.Text.Trim('.');

        this.data.DisplayText = account;

        var correct = await this.conversation
          .SayAsync("is that correct?")
          .WaitForCommandsAsync("yes", "no");

        done = (correct.Text == "yes");
      }
      this.accountNumber = account;
    }
    Task<ConversationResult> DisplaySayAsync(
      string displayText, string speechText = null)
    {
      this.data.DisplayText = displayText;

      var result = this.conversation.SayAsync(
        string.IsNullOrEmpty(speechText) ? displayText : speechText);

      return (result);
    }

    AudioRecorder audioRecorder;
    List<VerificationPhrase> verificationPhrases;
    OxfordSpeakerIdRestClient restClient;
    Conversation conversation;
    DisplayTextViewModel data;
    StorageFile recordingFile;
    VerificationProfile verificationProfile;
    string accountNumber;
    static readonly string RECORDING_FILE = "recording.bin";
    static readonly float DEFAULT_RECORD_TIME = 5.0f;
  }
}
